	TITLE "DCMOTOR SERVO CONTROL:  Revision:   1.9
;       Revised: 8/5/92
;       CREDIT:  Developed by Teknic Inc. 1992
;
;***************************************************************************** 

	PROCESSOR       PIC17C42
	LIST            COLUMNS=120, XREF=YES, NOWRAP,LINES=255

	#include "dcmotor.h17"

;***************************************************************************** 
;       global  variables
;
	CBLOCK  0x18
	DPX,DPX1,DPX2,DPX3              ; arithmetic accumulator
	AARG,AARG1,BARG,BARG1           ; multiply arguments
	ENDC

	CBLOCK  0x18
	TMP,TMP1,TMP2,TMP3              ; temporary variables
	MOVTMP,MOVTMP1,MOVTMP2,MOVTMP3  ; move temporary storage
	ENDC

	CBLOCK  0x20
	VL,VL1,VL2                      ; velocity limit
	AL,AL1,AL2                      ; acceleration limit
	
	KP,KP1                          ; proportional gain
	KV,KV1                          ; velocity gain
	KI,KI1                          ; integral gain
	IM                              ; integrator mode
	FV,FV1                          ; velocity feedforward
	FA,FA1                          ; acceleration feedforward
	
	VALBUF,VALBUF1,VALBUF2          ; iovalue buffer
	CVALBUF,CVALBUF1,CVALBUF2       ; iovalue buffer
	DVALBUF,DVALBUF1,DVALBUF2       ; iovalue buffer
	ISRBSR,ISRWREG                  ; isr save storage
	CMDCHAR,CMDTEMP,CMDPTRH,CMDPTRL ; command interface variables
	PARTEMP,PARLEN,PARPTR           ; parameter variables
	
	CPOSITION,CPOSITION1,CPOSITION2         ; shutter commanded position
	CVELOCITY,CVELOCITY1,CVELOCITY2         ; shutter commanded velocity
	CMPOSITION,CMPOSITION1,CMPOSITION2      ; shutter measured position
	CMVELOCITY,CMVELOCITY1,CMVELOCITY2      ; shutter measured velocity
	
	STRVALH,STRVALL                 ; string io variables
	HEXVAL,HEXTMP,HEXSTAT           ; hex io variables
	
	OPOSITION,OPOSITION1
	OPOSITION2,OPOSITION3           ; original commanded position
	POSITION,POSITION1,POSITION2    ; commanded position
	VELOCITY,VELOCITY1,VELOCITY2    ; commanded velocity

	NMOVVAL,NMOVVAL1,NMOVVAL2,NMOVVAL3      ; move value
	MOVVAL,MOVVAL1,MOVVAL2,MOVVAL3          ; move value
	HMOVVAL,HMOVVAL1,HMOVVAL2,HMOVVAL3      ; half move value
	MOVTIME,MOVTIME1                        ; move time in sample counts

	MOVSIGN                         ; 0x00 for positive, 0x80 for negative
	T1,T11                          ; time to maximum velocity
	T2,T21                          ; time for half the move
	TAU,TAU1                        ; total move time
	NMODE                           ; next move modetype
	MODE                            ; move modetype
	
	MPOSITION,MPOSITION1,MPOSITION2 ; measured position
	MVELOCITY,MVELOCITY1
	MVELOCITY2,MVELOCITY3           ; measured velocity
	POSERROR,POSERROR1,POSERROR2    ; position error
	VELERROR,VELERROR1,VELERROR2    ; velocity error

	SIGN                            ; multiply sign

	Y,Y1,Y2,Y3                      ; Y(k) before pwm conversion
	YPWM,YPWM1,YPWM2,YPWM3          ; pwm input
	YPWMIN,YPWMIN1,YPWMAX,YPWMAX1   ; pwm input limits

	U0,U01,U1,U11,U2,U21            ; saturated error at successive times

	SERVOFLAG                       ; servoflag = 0 => no servo
	MODETYPE                        ; mode flag(0=position,1=velocity,-1=torque)
	EXTSTAT                         ; external status register
	MOVSTAT                         ; move status register
	MOVFLAG                         ; move flag
	SATFLAG                         ; saturation flag (1=pos,-1=neg)
	INTEGRAL,INTEGRAL1              ; integrator

	C0,C01,C1,C11,C2,C21            ; difference equation coefficients

	DECVAL,DECSTAT,DECTMP,DECSIGN   ; decimal io variables

	A,A1,A2,A3                              ; commanded acceleration = +-AL, 0
	V,V1,V2,V3                              ; commanded velocity = +-VL, 0
	MOVPBUF,MOVPBUF1,MOVPBUF2,MOVPBUF3      ; commanded position buffer
	MOVVBUF,MOVVBUF1,MOVVBUF2,MOVVBUF3      ; commanded velocity buffer

	UPCOUNT,UPCOUNT1                ; running up counter
	DOWNCOUNT,DOWNCOUNT1            ; running down counter

	MOVDEL,MOVDEL1,MOVDEL2,MOVDEL3  ; move discretization delta
	PH2FLAT,PH2FLAT1                ; phase 2 flat iteration counter
	INDEXPOS,INDEXPOS1,INDEXPOS2    ; position at last index pulse

	SHIFTNUM                        ; # of bit shifts from middle 16

;**************************************************************************
;       For PICMASTER Debug/servo tuning Purposes Only

	CAPFLAG                         ; trace capture flag
	CAPCOUNT,CAPCOUNT1              ; PICMASTER trace capture counter
	CAPTMP,CAPTMP1                  ; trace capture temporary storage

;**************************************************************************

	ZERO,ONE                        ; constants

	ENDC

;***************************************************************************** 


	ORG     0x0                     ; reset vector
	GOTO    Startup                 ; startup vector

	ORG     0x20
	GOTO    InterruptPoll           ; interrupt vector


;***************************************************************************** 
; NAME:         Startup
;
; DESCRIPTION:  This routine is called on the hardware reset or when the 
;               program wishes to restore initial conditions. Initiali-
;               zation of run-time constants takes place here.
;
; RETURNS:      restart to safe and initial state
;
; STACK UTILIZATION: none
; TIMING (in cycles): X

Startup

	BSF     _glintd                 ; disable all interrupts
	AUTONO                          ; no auto increment or decrement
	MOVLW   0x18                    ; clear all memory locations [18,FF]
	MOVPF   WREG,FSR0

memloop
	CLRF    INDF0
	INCFSZ  FSR0
	GOTO    memloop
						    
	INCF    ONE
	
	MOVLB   BANK3                   ; BANK3 initialization
	MOVLW   TCON2_INIT
	MOVFP   WREG,TCON2

	MOVLW   PW1DCH_INIT             ; set duty cycle to midpoint
	MOVFP   WREG,PW1DCH
	MOVFP   WREG,PW2DCH

	MOVLW   PW1DCL_INIT
	MOVFP   WREG,PW1DCL
	MOVFP   WREG,PW2DCL
	
	MOVLW   TCON1_INIT              ; set organization of timers
	MOVFP   WREG,TCON1

	MOVLB   BANK2                   ; BANK2 initialization

	MOVLW   PR1_INIT
	MOVFP   WREG,PR1                ; initialize timer1 period

	MOVLW   PR2_INIT
	MOVFP   WREG,PR2                ; initialize timer2 period


	MOVLB   BANK0                   ; BANK0 initialization

	MOVLW   RTCSTA_INIT
	MOVFP   WREG,RTCSTA             ; sets RTC for external input

	MOVLW   RCSTA_INIT
	MOVFP   WREG,RCSTA              ; set receive status
	
	MOVLW   TXSTA_INIT              ; set transmit status
	MOVFP   WREG,TXSTA
	
	MOVLW   SPBRG_INIT              ; set baud rate
	MOVFP   WREG,SPBRG

	MOVLW   DDRB_INIT
	MOVFP   WREG,DDRB               ; set port B for whatever
		      
	MOVLB   BANK1                   ; BANK1 initialization

	MOVLW   0x08                    ; initialize some parameters
	MOVPF   WREG,KP+B1              ; proportional gain

	MOVLW   0x50
	MOVPF   WREG,KV+B1              ; derivative gain

	MOVLW   0x04
	MOVPF   WREG,KI+B1              ; integral gain

	MOVLW   0x01
	MOVPF   WREG,AL+B1              ; acceleration limit

	MOVLW   0xF0
	MOVPF   WREG,VL+B1              ; velocity limit

	BSF     SHIFTNUM,2              ; set shift number

	MOVPF   PW1DCH,YPWM+B1

	MOVLW   PWMAXL                  ; initialize pwm limits
	MOVPF   WREG,YPWMAX+B0
	MOVLW   PWMAXH
	MOVPF   WREG,YPWMAX+B1
	MOVLW   PWMINL
	MOVPF   WREG,YPWMIN+B0
	MOVLW   PWMINH
	MOVPF   WREG,YPWMIN+B1

	CLRF    PIR                     ; clear flags, set indiviual interrupts
	CLRF    INTSTA
	BSF     _tm2ie
	BSF     _peie

	BCF     _glintd                 ; enable interrupts

	MOVLB   BANK2

zeroctrs
	CLRF    RTCCL                   ; clear up counter
	CLRF    RTCCH

	CLRF    TMR3L                   ; clear down counter
	CLRF    TMR3H

	MOVLW   0xFF
delay   DECFSZ  WREG
	GOTO    delay

	MOVFP   RTCCL,WREG
	IORWF   RTCCH,W
	IORWF   TMR3L,W
	IORWF   TMR3H,W
	TSTFSZ  WREG
	GOTO    zeroctrs                ; motor still moving

	GOTO    PollingLoop

;***************************************************************************** 

;***************************************************************************** 
; NAME:         InterruptPoll


InterruptPoll

	MOVPF   BSR,ISRBSR              ; save BSR,WREG
	MOVPF   WREG,ISRWREG

	MOVLB   BANK1

	CALL    doMPosMVel              ; calculate measured position and
					; velocity

	CALL    doExtstat               ; evaluate external status

	RLNCF   MOVSTAT,W               ; if MOVFLAG=0 and MOVSTAT,BIT7=1
	ANDLW   0x01                    ; then do premove. This is only
	SUBWF   MOVFLAG,W               ; executed once at the beginning of 
	BTFSC   WREG,MSB                ; each move
	CALL    doPreMove

	BTFSC   MOVSTAT,BIT6            ; is motion continuing? 
	CALL    doMove                  ; if so, do move

	CALL    doError                 ; calculate position and velocity
					; error
	TSTFSZ  SERVOFLAG               ; test servoflag, if 0 then no servo
	CALL    doServo                 ; do servo

	TSTFSZ  CAPFLAG
	CALL    doCaptureRegs           ; for PIC-MASTER Trace Capture, demo purposes

	MOVLB   BANK1

	CLRF    PIR                     ; clear all interrupt request flags
	
	MOVFP   ISRBSR,BSR              ; restore BSR,WREG
	MOVFP   ISRWREG,WREG

	RETFIE

;*****************************************************************************

;***************************************************************************** 
; NAME:                 PollingLoop
; 
; DESCRIPTION:          The actual polling loop called after the board's 
;                       initialization
; 
; ENTRY CONDITIONS:     System globals and hardware initialized and the 
;                       interrupt processes started.
;

PollingLoop
	
	CALL    IdleFunction
	CALL    GetChk
	CPFSEQ  ONE                     ; GetChk, is receive buffer full?
	GOTO    PollingLoop

	CALL    GetChar                 ; if so, get character
	MOVPF   WREG,CMDCHAR            ; put in CMDCHAR
	GOTO    DoCommand

;*****************************************************************************

;***************************************************************************** 
; NAME:         IdleFunction

; DESCRIPTION:  This routine will perform work while doing waits in serial
;               I/O functions.
;

IdleFunction
	
	CLRWDT
	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         DoCommand
;
; DESCRIPTION:  Search command table for command and execute it.
;

DoCommand

	MOVLW   CMD_TABLE               ; CMD_TABLE LSB
	MOVPF   WREG,TBLPTRL
	MOVLW   page CMD_TABLE          ; CMD_TABLE MSB
	MOVPF   WREG,TBLPTRH

	TABLRD  1,1,CMDTEMP
tryNextCmd
	TABLRD  0,1,CMDTEMP             ; read entry from table
	TLRD    1,CMDPTRH
	TABLRD  0,1,CMDPTRL

	MOVFP   CMDTEMP,WREG
	CPFSLT  ZERO

	GOTO    noCommand               ; error if end of table

	CPFSEQ  CMDCHAR
	GOTO    tryNextCmd

	CALL    PutChar                 ; echo command

	MOVFP   CMDPTRH,PCLATH          ; indirect jump to command routine
	MOVFP   CMDPTRL,PCL
	NOP
cmdFinish
	CALL    PutChar                 ; send response character from
					; command routine followed by CR
	MOVLW   CR
	CALL    PutChar

	GOTO    PollingLoop

noCommand
	MOVLW   CMD_BAD                 ; send error character
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_null
; 
; DESCRIPTION:  The do nothing command used to determine if the chip is 
;               working. Initiated by a carriage return.
	
do_null
	MOVLW   CMD_OK
	GOTO    cmdFinish

;************************************************************************** 

;************************************************************************** 
; NAME:         do_move

; DESCRIPTION:  Commands the axis to move to a new position or velocity. 
;               Position data is relative, and in encoder counts. Velocity 
;               data is absolute, and in encoder counts/sample time multi-
;               plied by 256. All moves are performed by the controller such
;               that velocity and acceleration limits set into parameter 
;               memory will not be violated. All move commands are kept in a
;               one deep FIFO buffer. The command in the buffer is executed 
;               as soon as the currently executed command is complete.
;
; 
; ARGUMENTS:    M [800000,7FFFFF]
; 
do_move

	#if     DECIO

	CALL    GetDecVal

	#else

	CALL    GetVal

	#endif

	BTFSC   MOVSTAT,BIT7            ; test if buffer available
	GOTO    bufoverflow

	MOV24   VALBUF,NMOVVAL          ; if so, accept value into NMOVVAL and
	BSF     MOVSTAT,BIT7            ; set buffer full flag

	MOVLW   CMD_OK
	GOTO    cmdFinish

bufoverflow
	MOVLW   CMD_BAD                 ; else, return error
	GOTO    cmdFinish

;************************************************************************** 

;************************************************************************** 
; NAME:         do_mode
;
; DESCRIPTION:  An argument of "P" will cause all subsequent move commands
;               to beincremental position moves. A "V" argument will cause
;               all subsequent moves to be absolute velocity moves.
;
; ARGUMENTS:    O [P,V]
; 
	
do_mode
	
	CALL    IdleFunction            ; get single character loop
	CALL    GetChk
	CPFSEQ  ONE
	GOTO    do_mode
	CALL    GetChar
	MOVPF   WREG,STRVALL
	CLRF    MODETYPE                ; MODETYPE=0 for position moves
testP
	MOVLW   'P'                     ; position moves for type P
	CPFSEQ  STRVALL
	GOTO    testV
	GOTO    modeok
testV
	MOVLW   'V'                     ; velocity moves for type V
	CPFSEQ  STRVALL
	GOTO    testT
	INCF    MODETYPE                ; MODETYPE=1 for velocity moves
	GOTO    modeok
testT
	MOVLW   'T'                     ; TORQUE Moves for type 'T'
	CPFSEQ  STRVALL
	GOTO    modeerror
	SETF    MODETYPE                ; MODETYPE=-1 for torque moves
	CLRF    SERVOFLAG               ; disable servo
	GOTO    modeok
modeerror
	MOVLW   CMD_BAD                 ; mode error
	GOTO    cmdFinish
modeok
	MOVFP   STRVALL,WREG            ; echo type character
	CALL    PutChar

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_setparameter
;
; DESCRIPTION:  Sets controller parameters to the value given.
;
;       Parameter               #               Range
;
;       VL=velocity limit       0               [0,7FFFFF]
;       AL=acceleration limit   1               [0,7FFFFF]
;
;       KP=proportional gain    2               [8000,7FFF]
;       KP=velocity gain        3               [8000,7FFF]
;       KP=integral gain        4               [8000,7FFF]
;
;       IM=integrator mode      5               [0,3]
;
;       FV=velocity FF          6               [8000,7FFF] : Not Implemented
;       FA=acceleration FF      7               [8000,7FFF] : Not Implemented
;
; 
; ARGUMENTS:    S [0,FF] [800000,7FFFFF]
; 
	
do_setparameter
	
	CALL    GetPar                  ; get parameter number

	MOVLW   NUMPAR                  ; check if in range [0,NUMPAR]
	CPFSLT  VALBUF+B0
	GOTO    Serror

	MOVLW   PAR_TABLE               ; PAR_TABLE LSB
	MOVPF   WREG,TBLPTRL
	MOVLW   page PAR_TABLE          ; PAR_TABLE MSB
	MOVPF   WREG,TBLPTRH

	TABLRD  1,1,PARTEMP

setNextPar
	TLRD    1,PARTEMP               ; read entry from table
	TABLRD  0,1,PARLEN
	TABLRD  0,1,PARPTR

	MOVLW   NUMPAR                  ; error if end of table
	CPFSLT  PARTEMP
	GOTO    Serror

	MOVFP   PARTEMP,WREG
	CPFSEQ  VALBUF+B0
	GOTO    setNextPar

	MOVFP   PARPTR,WREG             ; pointer to parameter in FSR1
	MOVFP   WREG,FSR1

	#if     DECIO                   ; get new value in VALBUF

	CALL    GetDecVal

	#else

	CALL    GetVal

	#endif
		
	MOVLW   VALBUF                  ; pointer to VALBUF in FSR0
	MOVFP   WREG,FSR0
	AUTOINC                         ; set autoincrement
setGetMore
	MOVFP   INDF0,INDF1             ; move new value to parameter
	DECF    PARLEN
	TSTFSZ  PARLEN
	GOTO    setGetMore

	AUTONO                          ; no autoincrement


	MOVLW   CMD_OK
	GOTO    cmdFinish

Serror
	MOVLW   CMD_BAD
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_readparameter
; 
; DESCRIPTION:  Returns the present value of a parameter.
;
; ARGUMENTS:    R [0,FF]
; 
; RETURNS:      The present value of the requested parameter is returned.
	
do_readparameter
	
	CALL    GetPar                  ; get parameter number

	MOVLW   NUMPAR                  ; check if in range [0,NUMPAR]
	CPFSLT  VALBUF+B0
	GOTO    Rerror

	MOVLW   PAR_TABLE               ; PAR_TABLE LSB
	MOVPF   WREG,TBLPTRL
	MOVLW   page PAR_TABLE          ; PAR_TABLE MSB
	MOVPF   WREG,TBLPTRH

	TABLRD  1,1,PARTEMP

readNextPar
	TLRD    1,PARTEMP               ; read entry from table
	TABLRD  0,1,PARLEN
	TABLRD  0,1,PARPTR

	MOVLW   NUMPAR                  ; error if end of table
	CPFSLT  PARTEMP
	GOTO    Rerror

	MOVFP   PARTEMP,WREG
	CPFSEQ  VALBUF+B0
	GOTO    readNextPar

	MOVFP   PARPTR,WREG             ; pointer to parameter in FSR1
	MOVFP   WREG,FSR1

	MOVLW   VALBUF                  ; pointer to VALBUF in FSR1
	MOVFP   WREG,FSR0
	AUTOINC                         ; set autoincrement

	CLR24   VALBUF                  ; clear old VALBUF
readGetMore
	MOVFP   INDF1,INDF0             ; read parameter into VALBUF
	DECF    PARLEN
	TSTFSZ  PARLEN
	GOTO    readGetMore

	AUTONO                          ; no autoincrement


	#if     DECIO                   ; send parameter value

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif
		
	MOVLW   CMD_OK
	GOTO    cmdFinish

Rerror
	MOVLW   CMD_BAD
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_shutter
; 
; DESCRIPTION:  Returns the time (in sample time counts [0,FFFF]) since the 
;               start of the present move and captures the commanded and 
;               measured values of position and velocity at the time of the 
;               command.
;
;
; ARGUMENTS:    C
; 
; RETURNS:      The time since the start of the present move is returned.
; 
	
do_shutter

	BSF     _glintd                 ; disable all interrupts
	
	MOV24   POSITION,CPOSITION      ; capture commanded position
	MOV24   VELOCITY,CVELOCITY      ; capture commanded velocity
	MOV24   MPOSITION,CMPOSITION    ; capture measured position
	MOV24   MVELOCITY,CMVELOCITY    ; capture measured velocity

	CLRF    VALBUF+B2
	MOV16   MOVTIME,VALBUF          ; capture move time, move to VALBUF

	BCF     _glintd                 ; enable all interrupts

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_readcomposition
;
; DESCRIPTION:  Returns the commanded position count which was captured 
;               during the last shutter command.
;
; ARGUMENTS:    P
; 
; RETURNS:      The last captured position count is returned. [800000,7FFFFF]
; 
	
do_readcomposition
	
	MOV24   CPOSITION,VALBUF        ; move CPOSITION to VALBUF

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif
		
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_readcomvelocity
;
; DESCRIPTION:  Returns the commanded velocity multiplied by 256 which was 
;               captured during the last shutter command.
;
; ARGUMENTS:    V
; 
; RETURNS:      The last captured commanded velocity times 256 is returned.
;               [800000,7FFFFF]
; 
	
do_readcomvelocity
	
	MOV24   CVELOCITY,VALBUF        ; move commanded velocity to VALBUF

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif
		
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_readactposition
;
; DESCRIPTION:  Returns the measured position count which was captured during
;               the last shutter command.
;
; ARGUMENTS:    p
; 
; RETURNS:      The last captured measured position count is returned. 
;               [800000,7FFFFF]
; 
	
do_readactposition
	
	MOV24   CMPOSITION,VALBUF       ; move measured position to VALBUF

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_readactvelocity
;
; DESCRIPTION:  Returns the measured velocity multiplied by 256 which was 
;               captured during the last shutter command.
;
; ARGUMENTS:    v
; 
; RETURNS:      The last captured measured velocity times 256 is returned.
;               [800000,7FFFFF]
; 
	
do_readactvelocity
	
	MOV24   CMVELOCITY,VALBUF       ; move measured velocity to VALBUF

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif
		
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_externalstatus
; 
; DESCRIPTION:  Returns a two digit hex number which defines the state of 
;               the bits in the external status register. Issuing this 
;               command will clear all the bits in the external status 
;               register unless the event which set the bit is still true.
;
; 
; ARGUMENTS:    X
;
; RETURNS:      The external status register is returned. 
;
	
do_externalstatus
	
	BSF     _glintd
	MOVFP   EXTSTAT,WREG
	CLRF    EXTSTAT
	BCF     _glintd
	CALL    PutHex
					 
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME: do_movestatus
; 
; DESCRIPTION:  Returns a two digit hex number which defines the state of 
;               the bits in the move status register. Issuing this command 
;               will clear all the bits in the move status register unless 
;               the event which set the bit is still true.
;
; ARGUMENTS:    Y
; 
; RETURNS:      The move status register is returned.
; 
	
do_movestatus
	
	MOVFP   MOVSTAT,WREG
	CALL    PutHex
					 
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_readindposition
; 
; DESCRIPTION:  Returns the last index position captured in position counts.
;
; ARGUMENTS:    I
; 
; RETURNS:      The last captured index position is returned.
;
	
do_readindposition

	MOV24   INDEXPOS,VALBUF         ; move measured velocity to VALBUF

	#if     DECIO

	CALL    PutDecVal

	#else

	CALL    PutVal

	#endif

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_setposition
; 
; DESCRIPTION:  Sets the measured and commanded position to the value given. 
;               This command should not be sent unless the move FIFO buffer is empty.
;
; ARGUMENTS:    H [800000,7FFFFF]
; 
	
do_setposition
	
	#if     DECIO

	CALL    GetDecVal

	#else

	CALL    GetVal

	#endif
		
	MOV24   VALBUF,POSITION
	MOV24   VALBUF,MPOSITION

	CLR32   Y
     
	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;***************************************************************************** 
; NAME:         do_reset
; 
; DESCRIPTION:  Performs a software reset.
;
; ARGUMENTS:    Z
; 
	
do_reset
	
	MOVLW   CMD_OK
	CALL    PutChar
	GOTO    Startup


;***************************************************************************** 
; NAME:         do_stop
;
; DESCRIPTION:  Stops servo by clearing SERVOFLAG.
;
do_stop
	
	CLRF    SERVOFLAG

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 

;*****************************************************************************
; NAME:         do_capture
;
	
do_capture

	#if     DECIO

	CALL    GetDecVal

	#else

	CALL    GetVal

	#endif
		
	MOV16   VALBUF,CAPCOUNT
	MOV16   VALBUF,CAPTMP

	MOVLW   CMD_OK
	GOTO    cmdFinish

;***************************************************************************** 
; NAME:         GetChar
; 
; DESCRIPTION:  Get character from receive buffer.
;
GetChar

	MOVLB   BANK0                   ; set bank0
	MOVPF   RCREG,WREG              ; receive character

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         PutChar
;
; DESCRIPTION:  send character out the serial port
;
; ARGUMENTS:    WREG contains byte to be transmitted
; 

PutChar
	
	MOVLB   BANK1                   ; set bank1
bufwait                                 ; is transmit buffer empty?
	BTFSS   _tbmt
	GOTO    bufwait

	MOVLB   BANK0                   ; set bank0
shfwait
	BTFSS   _trmt                   ; is transmit shift register empty?
	GOTO    shfwait

	MOVPF   WREG,TXREG              ; if so, send character

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         GetChk
; 
; DESCRIPTION:  Check if character is in receive buffer.
;
	
GetChk
	MOVLB   BANK1                   ; set bank1
	MOVPF   PIR,WREG
	ANDLW   CHARREADY               ; return status in WREG
	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME: PutDec
; 
; DESCRIPTION:  Converts a hex value [0,F] in WREG to its ASCII equivalent.
;               The upper nibble of WREG is assumed to be zero.
;
; ENTRY CONDITIONS: WREG = value to be converted and sent in ASCII decimal
; 

	#if     DECIO

PutDec
	ADDLW   0x30                    ; convert to ASCII
	CALL    PutChar
	RETURN

	#endif

;***************************************************************************** 

;***************************************************************************** 
; NAME: PutHex
;
; DESCRIPTION:  Convert the WREG value to ASCII hexadecimal. The output 
;               format is two digits with the A-F parts in upper case and 
;               leading zeros. The result is sent out the serial port with 
;               PutChar.
; 
; ENTRY CONDITIONS: WREG = value to be converted and sent in ASCII hex
; 

PutHex
	
	MOVPF   WREG,HEXVAL
	SWAPF   WREG
	ANDLW   0x0F
	MOVPF   WREG,HEXTMP
	NEGW    WREG
	ADDLW   0x09
	BTFSS   WREG,MSB
	GOTO    puth20
	MOVLW   'A'-0x0A
	ADDWF   HEXTMP,W
	GOTO    puth25
puth20
	MOVLW   '0'
	ADDWF   HEXTMP,W
puth25
	CALL    PutChar

	MOVFP   HEXVAL,WREG
	ANDLW   0x0F
	MOVPF   WREG,HEXTMP
	NEGW    WREG
	ADDLW   0x09
	BTFSS   WREG,MSB
	GOTO    putl20
	MOVLW   'A'-0x0A
	ADDWF   HEXTMP,W
	GOTO    putl25
putl20
	MOVLW   '0'
	ADDWF   HEXTMP,W
putl25
	CALL    PutChar

	RETURN

;***************************************************************************** 

;*****************************************************************************
; NAME:         PutStr
; 
; DESCRIPTION:  Sends a character string out the serial port.
;

PutStr
	TABLRD  1,1,STRVALH
GetNextPair
	TLRD    1,STRVALH
	TABLRD  0,1,STRVALL

	MOVFP   STRVALH,WREG
	CPFSEQ  ZERO
	GOTO    putH
	RETURN
putH
	CALL    PutChar

	MOVFP   STRVALL,WREG
	CPFSEQ  ZERO
	GOTO    putL
	RETURN
putL
	CALL    PutChar

	GOTO    GetNextPair

;***************************************************************************** 

;***************************************************************************** 
; NAME:         GetHex
; 
; DESCRIPTION:  Receive an ASCII hex character from the serial port and
;               convert to numerical value.
;
; RETURNS:      numerical value in HEXVAL

GetHex

getnxt
	CALL    IdleFunction
	CALL    GetChk
	CPFSEQ  ONE
	GOTO    getnxt

	CLRF    HEXSTAT
	CALL    GetChar
	MOVPF   WREG,HEXVAL
	CALL    PutChar
	MOVLW   CR
	SUBWF   HEXVAL,W
	TSTFSZ  WREG
	GOTO    gth10
	GOTO    gthCR
gth10
		       
	MOVFP   HEXVAL,WREG
	SUBLW   '9'
	BTFSS   WREG,MSB
	GOTO    gth20

	MOVLW   0x09
	ADDWF   HEXVAL

gth20
	MOVLW   0x0F
	ANDWF   HEXVAL
	CLRF    HEXSTAT
	RETURN

gthCR   
	MOVLW   0x01
	MOVPF   WREG,HEXSTAT
	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         GetDec
; 
; DESCRIPTION:  Receive an ASCII decimal character from the serial port and
;               convert to its numerical value.
;
; ARGUMENTS:    numerical value is returned in DECVAL
; 

	#if     DECIO

GetDec
	
getdecnxt
	CALL    IdleFunction
	CALL    GetChk
	CPFSEQ  ONE
	GOTO    getdecnxt

	CALL    GetChar
	MOVPF   WREG,DECVAL
	CALL    PutChar

	MOVLW   CR
	SUBWF   DECVAL,W
	CPFSLT  ZERO
	GOTO    gtdCR
	MOVLW   MN
	SUBWF   DECVAL,W
	CPFSLT  ZERO
	GOTO    gtdMN
	MOVLW   SP
	SUBWF   DECVAL,W
	CPFSLT  ZERO
	GOTO    gtdSP
gtd09
	MOVLW   0x0F
	ANDWF   DECVAL
	CLRF    DECSTAT
	RETURN
gtdCR   
	MOVLW   DEC_CR
	MOVPF   WREG,DECSTAT
	RETURN
gtdMN   
	MOVLW   DEC_MN
	MOVPF   WREG,DECSTAT
	RETURN
gtdSP   
	MOVLW   DEC_SP
	MOVPF   WREG,DECSTAT
	RETURN

	#endif

;***************************************************************************** 

;***************************************************************************** 
; NAME:         getval
; 
; DESCRIPTION:  Get a value [800000,7FFFFF] from the serial port and place 
;               it in VALBUF.
;
	#if     DECIO
	#else

GetVal
	CLR24   VALBUF
getnext
	CALL    GetHex
	
	MOVLW   0x01
	CPFSEQ  HEXSTAT
	GOTO    shift
	RETURN
shift
	SWAPF   VALBUF+B2
	MOVFP   VALBUF+B2,WREG
	ANDLW   0xF0
	MOVPF   WREG,VALBUF+B2
	SWAPF   VALBUF+B1
	MOVFP   VALBUF+B1,WREG
	ANDLW   0x0F
	ADDWF   VALBUF+B2
	MOVFP   VALBUF+B1,WREG
	ANDLW   0xF0
	MOVPF   WREG,VALBUF+B1
	SWAPF   VALBUF+B0
	MOVFP   VALBUF+B0,WREG
	ANDLW   0x0F
	ADDWF   VALBUF+B1
	MOVFP   VALBUF+B0,WREG
	ANDLW   0xF0
	ADDWF   HEXVAL,W
	MOVPF   WREG,VALBUF+B0

	GOTO    getnext

	#endif

;*****************************************************************************

;***************************************************************************** 
; NAME:         GetDecVal
; 
; DESCRIPTION:  Get a value [-8388608,8388607] from the serial port and 
;               place it in VALBUF
;
; RETURNS:      numerical value is returned in VALBUF

	#if     DECIO

GetDecVal
	CLR24   VALBUF
	CALL    GetDec
	SETF    DECSIGN
	MOVLW   DEC_MN
	CPFSEQ  DECSTAT
	CLRF    DECSIGN

getdecnext
	CALL    GetDec
	
	MOVLW   DEC_CR
	CPFSEQ  DECSTAT
	GOTO    mul10
	GOTO    fixsign
mul10

	RLC24   VALBUF                  ; multiply VALBUF by two
	MOV24   VALBUF,DVALBUF          ; save in DVALBUF
	RLC24   VALBUF
	RLC24   VALBUF                  ; VALBUF now multiplied by eight
	ADD24   DVALBUF,VALBUF          ; VALBUF now multiplied by ten
	CLR24   DVALBUF
	MOVFP   DECVAL,WREG
	MOVPF   WREG,DVALBUF+B0
	ADD24   DVALBUF,VALBUF
	GOTO    getdecnext
fixsign
	CLRF    WREG
	CPFSGT  DECSIGN
	RETURN
	NEG24   VALBUF
	RETURN

	#endif

;***************************************************************************** 

;***************************************************************************** 
; NAME:         GetPar
; 
; DESCRIPTION:  Get a parameter number [0,FF] from the serial port and place 
;               it in VALBUF+B0.
;

GetPar
	
	CLR24   VALBUF

	CALL    GetHex
	MOVFP   HEXVAL,WREG
	ANDLW   0x0F
	MOVPF   WREG,VALBUF+B0
	SWAPF   VALBUF+B0
	
	CALL    GetHex
	MOVFP   VALBUF+B0,WREG

	ADDWF   HEXVAL,W
	MOVPF   WREG,VALBUF+B0

	RETURN

;*****************************************************************************

;***************************************************************************** 
; NAME:         PutVal
; 
; DESCRIPTION:  Sends the value in VALBUF [800000,7FFFFF] out the serial port.
;

	#if     DECIO
	#else

PutVal
	
	MOVFP   VALBUF+B2,WREG
	CALL    PutHex
	MOVFP   VALBUF+B1,WREG
	CALL    PutHex
	MOVFP   VALBUF+B0,WREG
	CALL    PutHex
					
	RETURN

	#endif

;*****************************************************************************

;***************************************************************************** 
; NAME:         PutDecVal
; 
; DESCRIPTION:  Send the value in VALBUF [-8388608,8388607] out the serial port.
;

	#if     DECIO

PutDecVal
	
	BTFSS   VALBUF+B2,MSB
	GOTO    pdpos
	NEG24   VALBUF
	MOVLW   MN
	CALL    PutChar
	GOTO    pddigits
pdpos
	MOVLW   SP
	CALL    PutChar

pddigits
	MOVLW   DEC_TABLE               ; DEC_TABLE LSB
	MOVPF   WREG,TBLPTRL
	MOVLW   page DEC_TABLE          ; DEC_TABLE MSB
	MOVPF   WREG,TBLPTRH

	TABLRD  0,1,DVALBUF+B0
readNextDec
	TLRD    0,DVALBUF+B0            ; read entry from table
	TABLRD  1,1,DVALBUF+B1
	TABLRD  0,1,DVALBUF+B2

	SETF    WREG                    ; unitsposition if end of table
	CPFSEQ  DVALBUF+B0
	GOTO    getdigit
	GOTO    unitsposition
getdigit
	INCF    DVALBUF+B0              ; restore to power of 10
	SETF    DECVAL                  ; set DECVAL to -1
inc
	INCF    DECVAL                  ; increment DECVAL
	SUB24   DVALBUF,VALBUF          ; check if in range
	BTFSS   VALBUF+B2,MSB
	GOTO    inc

	ADD24   DVALBUF,VALBUF          ; if so, correct VALBUF for next digit
	MOVFP   DECVAL,WREG             ; send DECVAL 
	CALL    PutDec

	GOTO    readNextDec             ; get next table entry

unitsposition
	MOVFP   VALBUF+B0,WREG          ; units position value now in VALBUF
	CALL    PutDec

	RETURN
	

	#endif

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doError
; 
; DESCRIPTION:  Calculates the position and velocity error.
;
	
doError

	MOV24   POSITION,POSERROR       ; calculate position error
	SUB24   MPOSITION,POSERROR

	BTFSC   POSERROR+B2,MSB         ; saturate error to lowest 16 bits
	GOTO    pneg
ppos
	MOVFP   POSERROR+B1,WREG
	ANDLW   0x80
	IORWF   POSERROR+B2
	CLRF    WREG
	CPFSGT  POSERROR+B2
	GOTO    psatok
	CLRF    POSERROR+B2             ; clear high byte for debug purposes
	MOVLW   0x7F
	MOVPF   WREG,POSERROR+B1
	SETF    POSERROR
	GOTO    psatok
pneg
	MOVFP   POSERROR+B1,WREG
	IORLW   0x7F
	ANDWF   POSERROR+B2
	SETF    WREG
	CPFSLT  POSERROR+B2
	GOTO    psatok
	SETF    POSERROR+B2             ; set high byte to 0xFF for debug purposes
	CLRF    POSERROR+B1
	BSF     POSERROR+B1,MSB
	CLRF    POSERROR
psatok

	MOV24   VELOCITY,VELERROR       ; calculate velocity error
	SUB24   MVELOCITY,VELERROR

	BTFSC   VELERROR+B2,MSB         ; saturate error to lowest 16 bits
	GOTO    vneg
vpos
	MOVFP   VELERROR+B1,WREG
	ANDLW   0x80
	IORWF   VELERROR+B2
	CLRF    WREG
	CPFSGT  VELERROR+B2
	GOTO    vsatok
	CLRF    VELERROR+B2
	MOVLW   0x7F
	MOVPF   WREG,VELERROR+B1
	SETF    VELERROR
	GOTO    vsatok
vneg
	MOVFP   VELERROR+B1,WREG
	IORLW   0x7F
	ANDWF   VELERROR+B2
	SETF    WREG
	CPFSLT  VELERROR+B2
	GOTO    vsatok
	SETF    VELERROR+B2
	CLRF    VELERROR+B1
	BSF     VELERROR+B1,MSB
	CLRF    VELERROR
vsatok
	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doServo
; 
; DESCRIPTION:  Performs the servo loop calculations.
;
doServo

	MOV16   POSERROR,U0             ; save new position error in U0

	LOADAB  U0,KP                   ; compute KP*U0
	CALL    Dmult
	MVPF32  DPX,Y                   ; Y=KP*U0

	CLRF    WREG                    ; if previous output saturated, do
	CPFSGT  SATFLAG                 ; not accumulate integrator
	CALL    doIntegral

	LOADAB  INTEGRAL,KI             ; compute KI*INTEGRAL
	CALL    Dmult
	ADD32   DPX,Y                   ; Y=KP*U0+KI*INTEGRAL

	MVFP16  U0,AARG                 ; compute KV*(U0-U1)
	SUB16   U1,AARG
	MVFP16  KV,BARG
	CALL    Dmult
	ADD32   DPX,Y                   ; Y=KP*U0+KI*INTEGRAL+KV*(U0-U1)

	CLRF    WREG
	CPFSGT  SHIFTNUM
	GOTO    grabok
	MOVFP   SHIFTNUM,TMP
grabloop
	RLC32   Y
	DECFSZ  TMP
	GOTO    grabloop
	
grabok
	CLRF    SATFLAG
	BTFSC   Y+B3,MSB                ; saturate to middle 16 bits,
	GOTO    negs                    ; keeping top 10 bits for PW1DCH 
poss                                    ; and PW1DCL
	MOVFP   Y+B2,WREG               ; check if Y >= 2**23
	ANDLW   0x80
	IORWF   Y+B3
	CLRF    WREG
	CPFSGT  Y+B3
	GOTO    zero6bits               ; if not, zero 6 bits

	INCF    SATFLAG                 ; if so, set Y=0x007FFFFF
	CLRF    Y+B3                      ; clear for debug purposes
	MOVLW   0x7F
	MOVPF   WREG,Y+B2
	SETF    Y+B1
	SETF    Y+B0
	GOTO    zero6bits
negs
	MOVFP   Y+B2,WREG               ; check if Y <= -2**23
	IORLW   0x7F
	ANDWF   Y+B3
	SETF    WREG
	CPFSLT  Y+B3
	GOTO    zero6bits               ; if not, zero 6 bits

	SETF    SATFLAG                 ; if so, set Y = 0xFF800000
	SETF    Y+B3
	CLRF    Y+B2
	BSF     Y+B2,MSB
	CLRF    Y+B1
	CLRF    Y+B0

zero6bits
	MOV24   Y+B1,YPWM+B0            ; move Y to YPWM and zero 6 bits
doTorque                                ; entry point for torque mode
	MOVLW   0xC0
	ANDWF   YPWM+B0

	BTFSC   YPWM+B1,MSB
	GOTO    tmlimit
tplimit
	BTFSS   EXTSTAT,BIT6
	GOTO    mplimitok
	CLR32   YPWM
	GOTO    mplimitok
tmlimit
	BTFSS   EXTSTAT,BIT5
	GOTO    mplimitok
	CLR32   YPWM
mplimitok
	MOVLW   PW1DCH_INIT             ; adjustment from bipolar to unipolar
	MOVPF   WREG,TMP+B1             ; for 50% duty cycle
	MOVLW   PW1DCL_INIT
	MOVPF   WREG,TMP+B0
	ADD16   TMP,YPWM

	CLRF    TMP+B1                  ; correct by 1 LSB
	MOVLW   0x40                    ; add one to bit5 of PW1DCL
	MOVPF   WREG,TMP+B0
	ADD16   TMP,YPWM

testmax
	CLRF    TMP+B2                  ; check pwm maximum limit
	CLRF    YPWM+B2                 ; LMD18200 must have a minimum pulse
	CLRF    YPWM+B3                 ; so duty cycle must not be 0 or 100%
	MVFP16  YPWMAX,TMP
	SUB24   YPWM,TMP
	BTFSS   TMP+B2,MSB
	GOTO    testmin
	MOV16   YPWMAX,YPWM             ; saturate to max
	GOTO    limitok
testmin
	CLRF    TMP+B2                  ; check pwm minimum limit
	CLRF    YPWM+B2
	CLRF    YPWM+B3
	MVFP16  YPWMIN,TMP
	SUB24   YPWM,TMP
	BTFSC   TMP+B2,MSB
	GOTO    limitok
	MOV16   YPWMIN,YPWM             ; saturate to min
limitok
	MOVLB   BANK3                   ; set new duty cycle
	MOVFP   YPWM+B0,PW1DCL
	MOVFP   YPWM+B1,PW1DCH

	MOV16   U0,U1                   ; push errors into U(k-1)

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doPreMove
; 
; DESCRIPTION:

doPreMove

	MOV24   NMOVVAL,MOVVAL          ; move buffer to MOVVAL
	BCF     MOVSTAT,BIT7            ; clear buffer flag
	BSF     MOVSTAT,BIT6            ; set motion status flag
	BSF     MOVSTAT,BIT5            ; set move in progress flag
	MOVFP   ONE,WREG
	MOVPF   WREG,MOVFLAG            ; initialize MOVEFLAG to 1


	CLRF    OPOSITION+B0            ; initialize buffers
	MOV24   POSITION,OPOSITION+B1
	MOV32   OPOSITION,MOVPBUF
	CLRF    SATFLAG
	CLR16   MOVTIME                 ; clear move times
	CLR16   T1                      ; 0 used as flag for no maximum speed
	CLR16   T2
	CLR16   TAU
	CLR32   MOVDEL                  ; clear move discretization error
	CLR16   PH2FLAT                 ; clear phase 2 flat counter

	TSTFSZ  MODETYPE
	GOTO    vmode
pmode
	MVFP24  MOVVAL,TMP
	BTFSS   TMP+B2,MSB
	GOTO    mvpos
	NEG24   TMP
mvpos
	CLRF    MOVTMP+B0               ; calculate  abs(MOVVAL) - 3
	CLRF    MOVTMP+B1               ; do immediate move if negative
	CLRF    MOVTMP+B2
	BSF     MOVTMP+B0,BIT0
	BSF     MOVTMP+B0,BIT1
	SUB24   MOVTMP,TMP

	BTFSS   TMP+B2,MSB                 ; check for zero move
	GOTO    nonzero
	SETF    SERVOFLAG               ; set servoflag to restore servo
	CLRF    MOVFLAG
	BCF     MOVSTAT,BIT5
	BCF     MOVSTAT,BIT6
	ADD24   MOVVAL,POSITION
	RETURN
nonzero
	CLR32   MOVVBUF

	MOVFP   MOVVAL+B2,WREG          ; move sign (00h=positive,80h=negative)
	ANDLW   0x80
	MOVPF   WREG,MOVSIGN

	CLRF    V+B3                    ; create appropriate velocity and
	MOV24   VL,V                    ; acceleration limits from move sign
	CLRF    A+B3
	MOV24   AL,A
	CLRF    WREG
	CPFSGT  MOVSIGN
	GOTO    minc
	NEG32   V
	NEG32   A
minc    
	CLRF    HMOVVAL+B0              ; evaluate MOVVAL/2
	MOV24   MOVVAL,HMOVVAL+B1
	RRC32   HMOVVAL                 ; half move in Q8
	GOTO    modeready

vmode
	BTFSC   MODETYPE,MSB            ; is it torque move?
	GOTO    tmode

	CLRF    HMOVVAL+B3              ; compute final minus initial velocity
	MOV24   MOVVAL,HMOVVAL
	BTFSC   MOVVAL+B2,MSB
	SETF    HMOVVAL+B3
	SUB32   MOVVBUF,HMOVVAL

	MOVFP   HMOVVAL+B3,WREG
	ANDLW   0x80
	MOVPF   WREG,MOVSIGN

	CLRF    V+B3                    ; create appropriate velocity and
	MOV24   VL,V                    ; acceleration limits from move sign
	CLRF    A+B3
	MOV24   AL,A
	CLRF    WREG
	CPFSGT  MOVSIGN
	GOTO    modeready
	NEG32   V
	NEG32   A

modeready
	CLRF    MOVVAL+B3
	BTFSC   MOVVAL+B2,MSB
	SETF    MOVVAL+B3

	SETF    SERVOFLAG               ; set servoflag to restore servo
					; if stopped
;*************************************************************************
;       For PICMASTER Debug/servo tuning puporses only Purposes Only
;
testCapCount
	MOVFP   CAPCOUNT+B0,WREG
	IORWF   CAPCOUNT+B1,W
	MOVPF   WREG,CAPFLAG
;*************************************************************************

	RETURN

tmode                                   ; torque/voltage mode
	MOV16   MOVVAL+B1,YPWM          ; set new commanded value
	CLRF    SERVOFLAG               ; disable servo
	CALL    doTorque                ; set pwm duty cycle
	CLRF    MOVFLAG
	BCF     MOVSTAT,BIT5
    goto    testCapCount

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doMove
; 
; DESCRIPTION:  In position mode, trapezoidal moves are performed. Phase1 
;               and phase2 respectively, are the periods for the first and 
;               second halves of the move. The move time is defined as zero 
;               at the beginning of the move,T2 is the time at half the move, T1 is the time when constant velocity
;               begins,(the region of constant velocity reduces to a point 
;               in the case where maximum speed is not realized, and the 
;               trapezoidal move degenerates into a trianglular move, 
;               together with T1=T2), and TAU is the total time of the move.  
;               The accelerations are +-AL or 0.
;
;
;                triangle speed                  trapezoidal speed
;
;                                                ________________
;                                               /                \
;                  /\                          /                  \
;                 /  \                        /                    \
;                /    \                      /                      \
;               /      \                    /                        \
;              /        \                  /                          \
;
;             0   T1=T2  TAU              0     T1      T2            TAU
;
;
;
;               Let x denote the undershoot and y the overshoot commanded 
;               at adjacent sample times as half the move is crossed.
;               In the case of a triangular move, the discretization error
;               is given by
;
;                        error = min (2x,2y)
;
;               For a trapezoidal move, the discretization error is 
;
;                       error = min (2x,y-x) <= .5*(maximum commanded speed)
;
;               This discretization error is resolved in the final sample 
;               time of the move by executing a step to the final position 
;               at zero speed. The method employed here the best possible 
;               performance with regard to discretization error without 
;               dynamically modifying velocity and acceleration limits.
;       
;
;
;               In velocity mode, ramp moves are performed.
;
;
;                                    / final velocity
;                                   /
;                                  /
;                                 /
;                                /
;             initial velocity  /
;
;                               0     TAU
;
;
	
doMove

	INC16   MOVTIME                 ; increment move time

	CALL    doPosVel                ; evaluate iterative equations

	TSTFSZ  MODETYPE
	GOTO    vmove
pmove
	MOVFP   ONE,WREG                ; test if in phase1
	CPFSEQ  MOVFLAG
	GOTO    phase2
phase1
	MVFP32  MOVDEL,MOVTMP           ; save previous discretization error
	MOV32   OPOSITION,MOVDEL        ; test if half move
	ADD32   HMOVVAL,MOVDEL
	SUB32   MOVPBUF,MOVDEL
	BTFSS   MOVSIGN,MSB
	GOTO    mpos1
	NEG32   MOVDEL
mpos1
	BTFSS   MOVDEL+B3,MSB
	GOTO    speedup                 ; continue to speed up if in phase1

	TFSZ16  T1                      ; if T1=0, maximum velocity not
					; reached, so T1=T2, otherwise T1
					; has been set in speedup
	GOTO    t2net1

	NEG32   A                       ; negate A for speeddown
	ADD32   MOVDEL,MOVTMP           ; test x-y < 0
	BTFSS   MOVTMP+B3,MSB           ; if new discretization error larger,
	GOTO    triok                   ; backup to define T2, otherwise ok

	SETF    T2+B0                   ; set T2=-1 for backup
	SETF    T2+B1
	NEG32   A                       ; negate A to undo
	CALL    undoPosVel
	NEG32   A                       ; negate A again for speeddown
	CALL    doPosVel                ; and reevaluate iterative equations
triok
	ADD16   MOVTIME,T2              ; add time to T2
	MOV16   T2,T1
	INCF    MOVFLAG                 ; increment move flag for phase2
	GOTO    mvok                    ; execute last phase1 move

t2net1
	SETF    T2+B0                   ; set T2=-1 for backup
	SETF    T2+B1
	ADD16   MOVTIME,T2              ; add time to T2

	MVFP32  MOVTMP,TMP              ; test if 3x-y < 0
	RLC32   MOVTMP
	ADD32   TMP,MOVTMP
	ADD32   MOVDEL,MOVTMP
	BTFSS   MOVTMP+B3,MSB           ; if new discretization error larger,
	GOTO    trapok                  ; take one more flat step
	SETF    PH2FLAT+B0
	SETF    PH2FLAT+B1
trapok
	ADD16   T2,PH2FLAT
	SUB16   T1,PH2FLAT
	INCF    MOVFLAG                 ; increment move flag for phase2
	GOTO    mvok                    ; execute last phase1 move

speedup
	MVFP32  V,MOVTMP                ; test if maximum velocity reached
	SUB32   MOVVBUF,MOVTMP
	BTFSS   MOVSIGN,MSB
	GOTO    mpos
	NEG32   MOVTMP
mpos
	BTFSS   MOVTMP+B3,MSB
	GOTO    mvok                    ; if not, execute move

	TFSZ16  T1                      ; if so, check to see if T1 has
					; already been set
	GOTO    mvok
	CALL    undoPosVel              ; if not, backup and redo iterative
	CLR32   A                       ; equations, resulting in an actual
	CALL    doPosVel                ; maximum speed <= VL
	SETF    T1+B0                   ; evaluate T1
	SETF    T1+B1
	ADD16   MOVTIME,T1
mvok
	MOV24   MOVPBUF+B1,POSITION     ; move Q8 calculated position to Q0 commanded position
	MOV24   MOVVBUF+B0,VELOCITY     ; move Q0 calculated velocity to Q0 commanded velocity
	RETURN


phase2
	TFSZ16  PH2FLAT                 ; is flat section finished?
	GOTO    flat

	TFSZ32  MOVVBUF                 ; is velocity zero?
	GOTO    mready                  ; if not, execute move

	CLRF    MOVFLAG                 ; if so, clear MOVFLAG
	BCF     MOVSTAT,BIT6            ; clear motion status flag
	BCF     MOVSTAT,BIT5            ; clear move in progress flag
	CLR32   A                       ; set zero velocity and acceleration,
	MOV16   MOVTIME,TAU
	MOV32   OPOSITION,MOVPBUF       ; execute last move to P(0)+MOVVAL
	ADD24   MOVVAL,MOVPBUF+B1
	GOTO    mready

flat
	SETF    MOVTMP+B0
	SETF    MOVTMP+B1
	ADD16   MOVTMP,PH2FLAT          ; decrement by one use DEC16

	TFSZ16  PH2FLAT
	GOTO    mready

	CLRF    A+B3                    ; begin speed down section
	MOV24   AL,A
	CLRF    WREG
	CPFSEQ  MOVSIGN
	GOTO    mready
	NEG32   A

mready
	MOV24   MOVPBUF+B1,POSITION
	MOV24   MOVVBUF+B0,VELOCITY
	RETURN

vmove
	MVFP32  MOVVAL,MOVTMP           ; test if final velocity reached
	SUB32   MOVVBUF,MOVTMP
	BTFSS   MOVSIGN,MSB
	GOTO    vmpos
	NEG32   MOVTMP
vmpos
	BTFSS   MOVTMP+B3,MSB
	GOTO    vmoveok                 ; if not, continue

	CLR32   A                       ; if so, set A=0 and continue with
	MOV32   MOVVAL,MOVVBUF          ; move unless the final velocity
					; is zero. 
	CLRF    MOVFLAG                 ; clear MOVFLAG
	BCF     MOVSTAT,BIT5            ; clear move in progress flag
	MOV16   MOVTIME,TAU
	TFSZ32  MOVVAL
	GOTO    vmoveok

	BCF     MOVSTAT,BIT6            ; if final velocity is zero, clear
					; motion status flag
vmoveok
	MOV24   MOVPBUF+B1,POSITION
	MOV24   MOVVBUF+B0,VELOCITY
	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doPosVel
; 
; DESCRIPTION:  Evaluates the iterative equations for trapezoidal move 
;               generation
;
;               V(k)=V(k-1)+A,          P(k)=P(k-1)+V(k-1)+A/2,
;
;               where abs(A)={AL,0} depending on the region of the trapezoid 
;               being executed.
;

doPosVel

	ADD32   MOVVBUF,MOVPBUF         ; P(k-1)+V(k-1)
	ADD32   A,MOVVBUF               ; V(k)=V(k-1)+A

	MVFP32  A,MOVTMP                ; compute A/2
	RRC32   MOVTMP

	ADD32   MOVTMP,MOVPBUF          ; P(k)=P(k-1)+V(k-1)+A/2,

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME: undoPosVel
; 
; DESCRIPTION:  Backward iteration of the equations for trapezoidal move 
;               generation
;
;               V(k-1)=V(k)-A,          P(k-1)=P(k)-V(k-1)-A/2,
;
;               where abs(A)={AL,0} depending on the region of the trapezoid 
;               being executed. This routine is used to reverse a step about 
;               to be made beyond a decision point.
;
	
undoPosVel

	SUB32   A,MOVVBUF               ; V(k-1)=V(k)-A
	SUB32   MOVVBUF,MOVPBUF         ; P(k)-V(k-1)

	MVFP32  A,MOVTMP                ; compute A/2
	RRC32   MOVTMP

	SUB32   MOVTMP,MOVPBUF          ; P(k-1)=P(k)-V(k-1)-A/2,

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         doMPosMVel
; 
; DESCRIPTION:  Calculates current position from UpCount and DownCount
;

doMPosMVel

; Do UpCounter first

	MVFP16  UPCOUNT,TMP+B0          ; save old upcount
readUp
	MOVPF   RTCCH,WREG
	MOVPF   RTCCL,UPCOUNT+B0
	CPFSEQ  RTCCH                   ; Skip next if HI hasn't changed
	GOTO    readUp                  ; HI changed, re-read LO
	MOVPF   WREG,UPCOUNT+B1         ; OK to store HI now

	CLRF    MVELOCITY+B0            ; clear bits below binary point

	MOV16   UPCOUNT,MVELOCITY+B1    ; compute upcount increment
	SUB16   TMP+B0,MVELOCITY+B1

; Now do DownCounter

	MVFP16  DOWNCOUNT,TMP+B0        ; save old downcount
readDown
	MOVLB   BANK2                   ; timers in Bank 2
	MOVPF   TMR3H,WREG
	MOVPF   TMR3L,DOWNCOUNT+B0
	CPFSEQ  TMR3H                   ; Skip next if HI hasn't changed
	GOTO    readDown                ; HI changed, re-read LO
	MOVPF   WREG,DOWNCOUNT+B1       ; OK to store HI now

	MVFP16  DOWNCOUNT+B0,TMP+B2     ; compute downcount increment
	SUB16   TMP+B0,TMP+B2

	SUB16   TMP+B2,MVELOCITY+B1     ; compute new measured velocity

	CLRF    MVELOCITY+B3            ; sign extend measured velocity for
	BTFSC   MVELOCITY+B2,MSB        ; 24 bit addition to measured position
	SETF    MVELOCITY+B3

	

	ADD24   MVELOCITY+B1,MPOSITION  ; compute new measured position
					; delta position = measured velocity

	RETURN

;***************************************************************************** 

;*****************************************************************************
; NAME:         doIntegral
; 
; DESCRIPTION:  Evaluates the integral for the servo calculations.
;
doIntegral

	ADD16   U0,INTEGRAL             ; do integral

	RETURN

;*****************************************************************************

;*****************************************************************************
; NAME:         doExtstat
;
; DESCRIPTION:  Get +limit,-limit,GPI from PORTB and set in EXTSTAT
;
doExtstat
	BTFSS   _intir
	GOTO    otherbits
	MOV24   MPOSITION,INDEXPOS
	BCF     _intir
	BSF     EXTSTAT,MSB
	
otherbits
	MOVLB   BANK0                   ; get +limit,-limit and GPI
	MOVFP   PORTB,WREG
	RRCF    WREG                    ; arrange in correct bit positions
	ANDLW   0x61
	MOVPF   WREG,TMP
	SWAPF   TMP
	IORWF   TMP,W
	IORWF   EXTSTAT                 ; set in EXTSTAT

	RETURN

;***************************************************************************** 

;***************************************************************************** 
; NAME:         Dmult
; 
; DESCRIPTION:  Mult: AARG (16 bits) * BARG (16 bits) -> DPX (32 bits)
;
;      (a) Load the 1st operand in locations AARG+B0 & AARG+B1 (16 bits)
;      (b) Load the 2nd operand in locations BARG+B0 & BARG+B1 (16 bits)
;      (c) CALL Dmult
;      (d) The 32 bit result is in locations (DPX+B0,DPX+B1,DPX+B2,DPX+B3)
;
;       In the signed case, a savings of 9 clks can be realized by choosing
;       BARG as the positive factor in the product when possible.
;       
; TIMING (worst case):  unsigned:                     173 clks
;                       signed:    if BARG positive:  170 clks
;                                  if BARG negative:  179 clks
;

;*************************************************************************

SIGNED  equ     TRUE            ; Set This To 'TRUE' for signed multiply
				; and 'FALSE' for unsigned.
;*************************************************************************
;               Multiplication Macro
;*************************************************************************
;
; TIMING:       unsigned:     11+7*10+8*11   = 169 clks 
;(worst case)   signed:       11+7*10+7*11+5 = 163 clks
;
MULTMAC  MACRO
	variable i

	  i = 0
	  #if   SIGNED
	     .while i < 15
	  #else
	     .while i < 16
	  #endif
	     .if i < 8
		BTFSC      BARG+B0,i        ; test low byte
		  .else
		BTFSC      BARG+B1,i-8      ; test high byte
	     .fi
		GOTO    add#v(i)
	     .if i < 8
		RLCF    DPX+B3,W        ; rotate sign into carry bit
		RRCF    DPX+B3          ; for i < 8, no meaningful bits
		RRCF    DPX+B2          ; are in DPX+B0
		RRCF    DPX+B1
	     .else
		RLCF    DPX+B3,W        ; rotate sign into carry bit
		RRCF    DPX+B3
		RRCF    DPX+B2
		RRCF    DPX+B1
		RRCF    DPX+B0
	     .fi
		i = i+1
	     .endw


		CLRF    DPX+B0          ; if we get here, BARG = 0
		RETURN



add0
		MOVFP   AARG+B0,WREG
		ADDWF   DPX+B2           ;add lsb
		MOVFP   AARG+B1,WREG
		ADDWFC  DPX+B3           ;add msb
		RLCF    AARG+B1,W       ; rotate sign into carry bit
		RRCF    DPX+B3          ; for i < 8, no meaningful bits
		RRCF    DPX+B2          ; are in DPX+B0
		RRCF    DPX+B1

	  i = 1
     
	  #if   SIGNED

		  .while i < 15
	  #else
	     .while i < 16
	  #endif
	     .if i < 8
		BTFSS      BARG+B0,i        ; test low byte
	     .else
		BTFSS      BARG+B1,i-8      ; test high byte
	     .fi
		GOTO    noadd#v(i)
add#v(i)
		MOVFP   AARG+B0,WREG
		ADDWF   DPX+B2           ;add lsb
		MOVFP   AARG+B1,WREG
		ADDWFC  DPX+B3           ;add msb

noadd#v(i)
	     .if i < 8

		RLCF    AARG+B1,W       ; rotate sign into carry bit
		RRCF    DPX+B3          ; for i < 8, no meaningful bits
		RRCF    DPX+B2          ; are in DPX+B0
		RRCF    DPX+B1

	     .else

		RLCF    AARG+B1,W       ; rotate sign into carry bit
		RRCF    DPX+B3
		RRCF    DPX+B2
		RRCF    DPX+B1
		RRCF    DPX+B0

	     .fi

	     i = i+1
	     .endw

	  #if   SIGNED

		RLCF    AARG+B1,W       ; since BARG is always made positive,
		RRCF    DPX+B3          ; the last bit is known to be zero.
		RRCF    DPX+B2
		RRCF    DPX+B1
		RRCF    DPX+B0

	  #endif

	ENDM

;          Double Precision Multiply ( 16x16 -> 32 )
;         ( AARG*BARG -> : 32 bit output in DPX
;
Dmult
     #if   SIGNED

	BTFSS   BARG+B1,MSB             ; test sign of BARG
	GOTO    argsok                  ; if positive, ok
	NEG16   AARG+B0                 ; if negative, then negate both
	NEG16   BARG+B0                 ; AARG and BARG

     #endif
argsok
	CLRF    DPX+B3                  ; clear initial partial product
	CLRF    DPX+B2

	MULTMAC                         ; use macro for multiplication

	RETURN

;***************************************************************************** 
;

;*****************************************************************************
; NAME:         doCaptureRegs
;
; DESCRIPTION:  Captures Desired Register Values To PIC-MASTER Trace Buffer
;               Intended for PICMASTER Demo/debug/servo tuning Purposes Only
;               Capture The following registers to Trace Buffer by putting
;               A Trace point on a TABLW instruction. Trace only 2nd Cycle
;
;               (a) POSERROR (position error : 16 bits)
;               (b) VELERROR (velocity error : 16 bits)
;               (c) MPOSITION (measured position value : 24 bits)
;               (d) MVELOCITY (measured velocity value : 24 bits)
;               (e) POSITION (commanded position : 24 bits)
;               (f) VELOCITY (commanded velocity : 24 bits)
;               (g) Y (output of servo loop : 32 bits)
;               (h) YPWM (output value written to PWM : 10 bits)
;
;
#define CaptureAddr     0x8000
;
doCaptureRegs
							; !end! hdr !skip start!
	movlw   (CaptureAddr & 0xff)
	movwf   TBLPTRL
	movlw   CaptureAddr/256
	movwf   TBLPTRH                                 ; setup table pointer address

	tablwt  0,0,POSERROR+B0 ; dummy tablwt
	tlwt    1,POSERROR+B1           ; now table latch = 16 bits contents of POSERROR
capPerr
	tablwt  0,0,POSERROR+B0 ; perform actual table write of POSERROR

	tablwt  0,0,VELERROR+B0
	tlwt    1,VELERROR+B1           ; capture Velocity error
capVerr
	tablwt  0,0,VELERROR+B0

	tablwt  0,0,MPOSITION+B0
	tlwt    1,MPOSITION+B1          ; capture measured position
capMpos
	tablwt  0,0,MPOSITION+B0

	tablwt  0,0,POSITION+B0
	tlwt    1,POSITION+B1           ; capture commanded position
capPos
	tablwt  0,0,POSITION+B0

	tablwt  0,0,MVELOCITY+B0
	tlwt    1,MVELOCITY+B1          ; capture measured velocity
capMvel
	tablwt  0,0,MVELOCITY+B0

	tablwt  0,0,VELOCITY+B0
	tlwt    1,VELOCITY+B1           ; capture commanded velocity
capVel
	tablwt  0,0,VELOCITY+B0

	DEC16   CAPTMP
	TFSZ16  CAPTMP
	RETURN
	CLRF    CAPFLAG
	MOV16   CAPCOUNT,CAPTMP
	HALT
	RETURN


;*****************************************************************************
;
;       
;       TABLES:
	

	CMD_START CMD_TABLE
	CMD_DEF do_null,DO_NULL
	CMD_DEF do_move,DO_MOVE
	CMD_DEF do_mode,DO_MODE
	CMD_DEF do_setparameter,DO_SETPARAMETER
	CMD_DEF do_readparameter,DO_READPARAMETER
	CMD_DEF do_shutter,DO_SHUTTER
	CMD_DEF do_readcomposition,DO_READCOMPOSITION
	CMD_DEF do_readcomvelocity,DO_READCOMVELOCITY
	CMD_DEF do_readactposition,DO_READACTPOSITION
	CMD_DEF do_readactvelocity,DO_READACTVELOCITY
	CMD_DEF do_externalstatus,DO_EXTERNALSTATUS
	CMD_DEF do_movestatus,DO_MOVESTATUS
	CMD_DEF do_readindposition,DO_READINDPOSITION
	CMD_DEF do_setposition,DO_SETPOSITION
	CMD_DEF do_reset,DO_RESET
	CMD_DEF do_stop,DO_STOP
	CMD_DEF do_capture,DO_CAPTURE
	CMD_END                                                    


PAR_TABLE       DATA    0x0003
		DATA    VL
		DATA    0x0103
		DATA    AL
		DATA    0x0202
		DATA    KP
		DATA    0x0302
		DATA    KV
		DATA    0x0402
		DATA    KI
		DATA    0x0501
		DATA    IM
		DATA    0x0602
		DATA    FV
		DATA    0x0702
		DATA    FA
		DATA    NUMPAR

	#if     DECIO

DEC_TABLE       DATA    0x423F
		DATA    0x000F
		DATA    0x869F
		DATA    0x0001
		DATA    0x270F
		DATA    0x0000
		DATA    0x03E7
		DATA    0x0000
		DATA    0x0063
		DATA    0x0000
		DATA    0x0009
		DATA    0x0000
		DATA    0xFFFF
	#endif

	
	
	
	END
